/* Fo
 * fo-doc-pdflib.c: Wrapper for libfo output document
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-doc-pdflib.c,v 1.6 2004/06/03 21:19:21 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include <pango/pangopdflib.h>
#include "libfo/fo-utils.h"
#include "area/fo-area-layout.h"
#include "libfo/fo-doc-pdflib-private.h"
#include "libfo/fo-doc-commands.h"
#include "libfo/fo-layout-pdflib-private.h"
#include "datatype/fo-color.h"

const char *fo_doc_pdflib_error_messages [] = {
  N_("FoDoc error"),
  N_("Cannot open output document: '%s'")
};

static void fo_doc_pdflib_init          (FoDocPDFLIB *object);
static void fo_doc_pdflib_base_init     (FoDocPDFLIBClass *klass);
static void fo_doc_pdflib_base_finalize (FoDocPDFLIBClass *klass);
static void fo_doc_pdflib_class_init    (FoDocPDFLIBClass *klass);
static void fo_doc_pdflib_finalize      (GObject    *object);
static void* fo_doc_pdflib_allocproc    (PDF        *pdf_doc,
					 size_t      size,
					 const char *caller);
static void* fo_doc_pdflib_reallocproc  (PDF        *pdf_doc,
					 void       *mem,
					 size_t      size,
					 const char *caller);
static void  fo_doc_pdflib_freeproc     (PDF        *pdf_doc,
					 void       *mem);

static FoLayout *    fo_doc_pdflib_get_new_layout   (FoDoc        *fo_doc);

static void          fo_doc_pdflib_begin_page       (FoDoc        *fo_doc,
						     gfloat        width,
						     gfloat        height);
static void          fo_doc_pdflib_end_page         (FoDoc        *fo_doc);
static FoDatatype *  fo_doc_pdflib_get_fill_color   (FoDoc        *fo_doc);
static void          fo_doc_pdflib_set_fill_color   (FoDoc        *fo_doc,
						     FoDatatype   *color);
static FoDatatype *  fo_doc_pdflib_get_stroke_color (FoDoc        *fo_doc);
static void          fo_doc_pdflib_set_stroke_color (FoDoc        *fo_doc,
						     FoDatatype   *color);
static void          fo_doc_pdflib_translate        (FoDoc        *fo_doc,
						     gfloat        x,
						     gfloat        y);
static FoDocLineCap  fo_doc_pdflib_get_line_cap     (FoDoc        *fo_doc);
static void          fo_doc_pdflib_set_line_cap     (FoDoc        *fo_doc,
						     FoDocLineCap  line_cap);
static FoDocLineJoin fo_doc_pdflib_get_line_join    (FoDoc        *fo_doc);
static void          fo_doc_pdflib_set_line_join    (FoDoc        *fo_doc,
						     FoDocLineJoin line_join);
static gfloat        fo_doc_pdflib_get_line_width   (FoDoc        *fo_doc);
static void          fo_doc_pdflib_set_line_width   (FoDoc        *fo_doc,
						     gfloat        line_width);
static void          fo_doc_pdflib_set_dash         (FoDoc        *fo_doc,
						     gfloat        b,
						     gfloat        w);
static void          fo_doc_pdflib_translate        (FoDoc        *fo_doc,
						     gfloat        x,
						     gfloat        y);
static void          fo_doc_pdflib_clip             (FoDoc        *fo_doc);
static void          fo_doc_pdflib_save             (FoDoc        *fo_doc);
static void          fo_doc_pdflib_restore          (FoDoc        *fo_doc);
static void          fo_doc_pdflib_line_to          (FoDoc        *fo_doc,
						     gfloat        x,
						     gfloat        y);
static void          fo_doc_pdflib_move_to          (FoDoc        *fo_doc,
						     gfloat        x,
						     gfloat        y);
static void          fo_doc_pdflib_line_stroked     (FoDoc        *fo_doc,
						     gfloat        x0,
						     gfloat        y0,
						     gfloat        x1,
						     gfloat        y1);
static void          fo_doc_pdflib_rect_stroked     (FoDoc        *fo_doc,
						     gfloat        x,
						     gfloat        y,
						     gfloat        width,
						     gfloat        height);
static void          fo_doc_pdflib_rect_filled      (FoDoc        *fo_doc,
						     gfloat        x,
						     gfloat        y,
						     gfloat        width,
						     gfloat        height);
static void          fo_doc_pdflib_fill             (FoDoc        *fo_doc);
static void          fo_doc_pdflib_stroke           (FoDoc        *fo_doc);
static void          fo_doc_pdflib_fill_stroke      (FoDoc        *fo_doc);
static void          fo_doc_pdflib_place_image      (FoDoc        *fo_doc,
						     gint          image,
						     gfloat        x,
						     gfloat        y,
						     gfloat        scale);
static gint          fo_doc_pdflib_open_image_file  (FoDoc        *fo_doc,
						     const char   *imagetype,
						     const char   *filename,
						     const char   *stringparam,
						     gint          intparam);
static void          fo_doc_pdflib_close_image      (FoDoc        *fo_doc,
						     gint          image);
static void          fo_doc_pdflib_render_layout_lines (FoDoc   *fo_doc,
							FoArea  *area_layout,
							gfloat   x,
							gfloat   y);
static void 	     fo_doc_pdflib_render_layout       (FoDoc   *fo_doc,
							FoArea  *area_layout,
							gfloat   x,
							gfloat   y);

static void          fo_doc_pdflib_debug_dump (FoObject *object,
					       gint      depth);

static gpointer parent_class;

/**
 * fo_doc_pdflib_error_quark:
 * 
 * Get the error quark for #FoDoc.
 *
 * If the quark does not yet exist, create it.
 * 
 * Return value: Quark associated with #FoDoc errors.
 **/
GQuark
fo_doc_pdflib_error_quark (void)
{
  static GQuark quark = 0;
  if (quark == 0)
    quark = g_quark_from_static_string ("FoDoc error");
  return quark;
}


/**
 * fo_doc_pdflib_get_type:
 * 
 * Register the #FoDoc object type.
 * 
 * Return value: #GType value of the #FoDoc object type.
 **/
GType
fo_doc_pdflib_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (FoDocPDFLIBClass),
        (GBaseInitFunc) fo_doc_pdflib_base_init,
        (GBaseFinalizeFunc) fo_doc_pdflib_base_finalize,
        (GClassInitFunc) fo_doc_pdflib_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (FoDocPDFLIB),
        0,              /* n_preallocs */
        (GInstanceInitFunc) fo_doc_pdflib_init,
	NULL
      };
      
      object_type = g_type_register_static (FO_TYPE_DOC,
                                            "FoDocPDFLIB",
                                            &object_info,
					    0);
    }
  
  return object_type;
}

/**
 * fo_doc_pdflib_init:
 * @fo_doc_pdflib: #FoDocPDFLIB object to initialise.
 * 
 * Implements #GInstanceInitFunc for #FoDocPDFLIB.
 **/
void
fo_doc_pdflib_init (FoDocPDFLIB *object)
{
  PangoLanguage *pango_lang = pango_language_from_string ("EN");

  object->pdf_doc = PDF_new2 (NULL,
			      &fo_doc_pdflib_allocproc,
			      &fo_doc_pdflib_reallocproc,
			      &fo_doc_pdflib_freeproc,
			      &object);
  FO_DOC (object)->pango_context = pango_pdflib_get_context (300,
							     300);

  pango_context_set_base_dir (FO_DOC (object)->pango_context,
			      PANGO_DIRECTION_LTR);
  pango_context_set_language (FO_DOC (object)->pango_context,
			      pango_lang);
}

/**
 * fo_doc_pdflib_base_init:
 * @klass: #FoDocPDFLIBClass base class object to initialise
 * 
 * Implements #GBaseInitFunc for #FoDocPDFLIBClass
 **/
void
fo_doc_pdflib_base_init (FoDocPDFLIBClass *klass)
{
  FoDocClass *fo_doc_class = FO_DOC_CLASS (klass);

  PDF_boot ();

  fo_doc_class->open_file           = fo_doc_pdflib_open_file;

  fo_doc_class->get_new_layout      = fo_doc_pdflib_get_new_layout;

  fo_doc_class->begin_page          = fo_doc_pdflib_begin_page;
  fo_doc_class->end_page            = fo_doc_pdflib_end_page;
  fo_doc_class->get_fill_color      = fo_doc_pdflib_get_fill_color;
  fo_doc_class->set_fill_color      = fo_doc_pdflib_set_fill_color;
  fo_doc_class->get_stroke_color    = fo_doc_pdflib_get_stroke_color;
  fo_doc_class->set_stroke_color    = fo_doc_pdflib_set_stroke_color;
  fo_doc_class->translate           = fo_doc_pdflib_translate;
  fo_doc_class->get_line_cap        = fo_doc_pdflib_get_line_cap;
  fo_doc_class->set_line_cap        = fo_doc_pdflib_set_line_cap;
  fo_doc_class->get_line_join       = fo_doc_pdflib_get_line_join;
  fo_doc_class->set_line_join       = fo_doc_pdflib_set_line_join;
  fo_doc_class->get_line_width      = fo_doc_pdflib_get_line_width;
  fo_doc_class->set_line_width      = fo_doc_pdflib_set_line_width;
  fo_doc_class->set_dash            = fo_doc_pdflib_set_dash;
  fo_doc_class->translate           = fo_doc_pdflib_translate;
  fo_doc_class->clip                = fo_doc_pdflib_clip;
  fo_doc_class->save                = fo_doc_pdflib_save;
  fo_doc_class->restore             = fo_doc_pdflib_restore;
  fo_doc_class->line_to             = fo_doc_pdflib_line_to;
  fo_doc_class->move_to             = fo_doc_pdflib_move_to;
  fo_doc_class->line_stroked        = fo_doc_pdflib_line_stroked;
  fo_doc_class->rect_stroked        = fo_doc_pdflib_rect_stroked;
  fo_doc_class->rect_filled         = fo_doc_pdflib_rect_filled;
  fo_doc_class->fill                = fo_doc_pdflib_fill;
  fo_doc_class->stroke              = fo_doc_pdflib_stroke;
  fo_doc_class->fill_stroke         = fo_doc_pdflib_fill_stroke;
  fo_doc_class->place_image         = fo_doc_pdflib_place_image;
  fo_doc_class->open_image_file     = fo_doc_pdflib_open_image_file;
  fo_doc_class->close_image         = fo_doc_pdflib_close_image;
  fo_doc_class->render_layout_lines = fo_doc_pdflib_render_layout_lines;
  fo_doc_class->render_layout       = fo_doc_pdflib_render_layout;
}

/**
 * fo_doc_pdflib_base_finalize:
 * @klass: #FoDocPDFLIB object class to finalise.
 * 
 * Implements #GBaseFinalizeFunc for #FoDocPDFLIB.
 **/
void
fo_doc_pdflib_base_finalize (FoDocPDFLIBClass *klass)
{
  PDF_shutdown ();
}

/**
 * fo_doc_pdflib_class_init:
 * @klass: #FoDocPDFLIBClass object to initialise.
 * 
 * Implements #GClassInitFunc for #FoDocPDFLIBClass.
 **/
void
fo_doc_pdflib_class_init (FoDocPDFLIBClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  FoObjectClass *fo_object_class = FO_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize = fo_doc_pdflib_finalize;

  fo_object_class->debug_dump = fo_doc_pdflib_debug_dump;
}

/**
 * fo_doc_pdflib_finalize:
 * @object: #FoDocPDFLIB object to finalize.
 * 
 * Implements #GObjectFinalizeFunc for #FoDocPDFLIB.
 **/
void
fo_doc_pdflib_finalize (GObject *object)
{
  FoDocPDFLIB *fo_doc;

  fo_doc = FO_DOC_PDFLIB (object);

  PDF_close (fo_doc->pdf_doc);
  PDF_delete (fo_doc->pdf_doc);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * fo_doc_pdflib_new:
 * 
 * Creates a new #FoDoc.
 * 
 * Return value: the newly created #FoDoc.
 **/
FoDoc *
fo_doc_pdflib_new (void)
{
  return FO_DOC (g_object_new (fo_doc_pdflib_get_type (),
			       NULL));
}

/**
 * fo_doc_pdflib_allocproc:
 * @pdf_doc: PDFlib PDF object, may be NULL.
 * @size:    Amount of memory to allocate.
 * @caller:  PDFlib function name.
 * 
 * Allocate @size bytes of memory. If @size is 0, it returns NULL.
 * 
 * Return value: Pointer to the allocated memory.
 **/
static void*
fo_doc_pdflib_allocproc (PDF *pdf_doc,
			 size_t size,
			 const char *caller)
{
  return g_malloc ((gulong) size);
}

/**
 * fo_doc_pdflib_reallocproc:
 * @pdf_doc: PDFlib PDF object.
 * @mem:     The memory to reallocate.
 * @size:    New size of the memory in bytes.
 * @caller:  PDFlib function name.
 * 
 * Reallocates the memory pointed to by @mem.
 * 
 * Return value: The new address of the allocated memory.
 **/
static void*
fo_doc_pdflib_reallocproc (PDF *pdf_doc,
			   void *mem,
			   size_t size,
			   const char *caller)
{
  return g_realloc (mem, (gulong) size);
}

/**
 * fo_doc_pdflib_freeproc:
 * @pdf_doc: PDFlib PDF object.
 * @mem:     The memory to free.
 * 
 * Frees the memory pointed to by @mem.
 **/
static void
fo_doc_pdflib_freeproc (PDF *pdf_doc,
			void *mem)
{
  g_free (mem);
}

/**
 * fo_doc_pdflib_get_pdf_doc:
 * @fo_doc: #FoDoc
 * 
 * Get the PDFlib document in @fo_doc.
 * 
 * Return value: Pointer to PDFlib document.
 **/
PDF*
fo_doc_pdflib_get_pdf_doc (FoDoc *fo_doc)
{
  g_return_val_if_fail (fo_doc != NULL, NULL);

  return FO_DOC_PDFLIB (fo_doc)->pdf_doc;
}

/**
 * fo_doc_pdflib_set_parameter:
 * @fo_doc: #FoDoc.
 * @key:    Name of string parameter.
 * @value:  String parameter value.
 * 
 * Set the parameter of @fo_doc named @key with string type to @value.
 **/
void
fo_doc_pdflib_set_parameter (FoDoc *fo_doc,
			     const gchar *key,
			     const gchar *value)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_set_parameter (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
		     key,
		     value);
}

/**
 * fo_doc_pdflib_set_value:
 * @fo_doc: #FoDoc.
 * @key:    Name of numeric parameter.
 * @value:  Numeric parameter value.
 * 
 * Set the parameter of @fo_doc named @key with numeric type to @value.
 **/
void
fo_doc_pdflib_set_value (FoDoc       *fo_doc,
			 const gchar *key,
			 gfloat       value)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_set_value (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
		 key,
		 value);
}

/**
 * fo_doc_pdflib_set_info:
 * @fo_doc: #FoDoc.
 * @key:    Document information item name.
 * @value:  Document information item value.
 * 
 * Set the document information field of @fo_doc named @key to @value.
 **/
void
fo_doc_pdflib_set_info (FoDoc       *fo_doc,
			const gchar *key,
			const gchar *value)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_set_info (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
		key,
		value);
}

/**
 * fo_doc_pdflib_open_file:
 * @fo_doc:   #FoDoc.
 * @filename: File to open as output.
 * @error:    #GError that is set if an error occurs.
 * 
 * Open @filename as the output file for @fo_doc.
 **/
void
fo_doc_pdflib_open_file (FoDoc       *fo_doc,
			 const gchar *filename,
			 GError     **error)
{
  gint pdflib_result;

  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);
  g_return_if_fail (filename != NULL);
  g_return_if_fail (error == NULL || *error == NULL);

  pdflib_result = PDF_open_file (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
				 filename);

  if (pdflib_result == -1)
    {
      g_set_error (error,
		   FO_DOC_ERROR,
		   FO_DOC_ERROR_OPEN_FAILED,
		   fo_doc_error_messages[FO_DOC_ERROR_OPEN_FAILED],
		   filename);
    }
}

/**
 * fo_doc_pdflib_get_new_layout:
 * @fo_doc: #FoDoc.
 * 
 * Get a new #FoLayout for use with @fo_doc.
 *
 * Return value: New #FoLayout.
 **/
FoLayout *
fo_doc_pdflib_get_new_layout (FoDoc *fo_doc)
{
  FoLayout *fo_layout;

  g_return_val_if_fail (FO_IS_DOC_PDFLIB (fo_doc), NULL);

  fo_layout = fo_layout_pdflib_new ();

  fo_layout->fo_doc = fo_doc;
  fo_layout->pango_layout =
    pango_layout_new (fo_doc->pango_context);

  return fo_layout;
}

/**
 * fo_doc_pdflib_begin_page:
 * @fo_doc: #FoDoc.
 * @width:  Width of the new page.
 * @height: Height of the new page.
 * 
 * Add a new page to @fo_doc.
 *
 * This must always be paired with a matching #fo_doc_pdflib_end_page call.
 **/
void
fo_doc_pdflib_begin_page (FoDoc *fo_doc,
			  gfloat width,
			  gfloat height)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_begin_page (FO_DOC_PDFLIB (fo_doc)->pdf_doc, width, height);
}

/**
 * fo_doc_pdflib_end_page:
 * @fo_doc: #FoDoc.
 * 
 * Finish the current page of @fo_doc.
 **/
void
fo_doc_pdflib_end_page (FoDoc *fo_doc)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_end_page (FO_DOC_PDFLIB (fo_doc)->pdf_doc);
}

/**
 * fo_doc_pdflib_get_fill_color:
 * @fo_doc: #FoDoc.
 * 
 * Get the current fill color of @fo_doc.
 * 
 * Return value: #FoDatatype for the current fill color.
 **/
FoDatatype*
fo_doc_pdflib_get_fill_color (FoDoc *fo_doc)
{
  g_return_val_if_fail (FO_IS_DOC_PDFLIB (fo_doc), NULL);

  return fo_doc->fill_color;
}

/**
 * fo_doc_pdflib_set_fill_color:
 * @fo_doc: #FoDoc.
 * @color:  #FoDatatype for new fill color.
 * 
 * Set the fill color of @fo_doc to @color.
 **/
void
fo_doc_pdflib_set_fill_color (FoDoc *fo_doc,
			      FoDatatype *color)
{
  guint16 red;
  guint16 green;
  guint16 blue;

  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);
  g_return_if_fail (FO_IS_COLOR (color));

  red = fo_color_get_red (color);
  green = fo_color_get_green (color);
  blue = fo_color_get_blue (color);

  if (color != NULL)
    {
      g_object_ref (color);
    }
  if (fo_doc->fill_color != NULL)
    {
      g_object_unref (fo_doc->fill_color);
    }
  fo_doc->fill_color = color;

  PDF_setcolor (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
		"fill",
		"rgb",
		(float) red / FO_COLOR_COMPONENT_MAX,
		(float) green / FO_COLOR_COMPONENT_MAX,
		(float) blue / FO_COLOR_COMPONENT_MAX,
		0.0);
}

/**
 * fo_doc_pdflib_get_stroke_color:
 * @fo_doc: #FoDoc.
 * 
 * Get the current stroke color of @fo_doc.
 * 
 * Return value: #FoDatatype for the current stroke color.
 **/
FoDatatype*
fo_doc_pdflib_get_stroke_color (FoDoc *fo_doc)
{
  g_return_val_if_fail (FO_IS_DOC_PDFLIB (fo_doc), NULL);

  return fo_doc->stroke_color;
}

/**
 * fo_doc_pdflib_set_stroke_color:
 * @fo_doc: #FoDoc.
 * @color:  #FoDatatype for new stroke color.
 * 
 * Set the stroke color of @fo_doc to @color.
 **/
void
fo_doc_pdflib_set_stroke_color (FoDoc *fo_doc,
				FoDatatype *color)
{
  guint16 red;
  guint16 green;
  guint16 blue;

  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);
  g_return_if_fail (FO_IS_COLOR (color));

  red = fo_color_get_red (color);
  green = fo_color_get_green (color);
  blue = fo_color_get_blue (color);

  if (color != NULL)
    {
      g_object_ref (color);
    }
  if (fo_doc->stroke_color != NULL)
    {
      g_object_unref (fo_doc->stroke_color);
    }
  fo_doc->stroke_color = color;

  PDF_setcolor (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
		"stroke",
		"rgb",
		(float) red / FO_COLOR_COMPONENT_MAX,
		(float) green / FO_COLOR_COMPONENT_MAX,
		(float) blue / FO_COLOR_COMPONENT_MAX,
		0.0);
}

/**
 * fo_doc_pdflib_get_line_cap:
 * @fo_doc: #FoDoc.
 * 
 * Get the current "linecap" parameter value of @fo_doc.
 * 
 * Return value: Current "linecap" parameter value.
 **/
FoDocLineCap
fo_doc_pdflib_get_line_cap (FoDoc        *fo_doc)
{
  g_return_val_if_fail (FO_IS_DOC_PDFLIB (fo_doc), FO_DOC_LINE_CAP_INVALID);

  return fo_doc->line_cap;
}

/**
 * fo_doc_pdflib_set_line_cap:
 * @fo_doc:   #FoDoc.
 * @line_cap: New "linecap" parameter value.
 * 
 * Set the "linecap" parameter value of @fo_doc.
 **/
void
fo_doc_pdflib_set_line_cap (FoDoc       *fo_doc,
			    FoDocLineCap line_cap)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);
  g_return_if_fail (line_cap < FO_DOC_LINE_CAP_LIMIT);

  /*if (line_cap != fo_doc->line_cap)
    {*/
      fo_doc->line_cap = line_cap;
      PDF_setlinecap (FO_DOC_PDFLIB (fo_doc)->pdf_doc, line_cap);
      /*}*/
}

/**
 * fo_doc_pdflib_get_line_join:
 * @fo_doc: #FoDoc.
 * 
 * Get the current "linejoin" parameter value of @fo_doc.
 * 
 * Return value: Current "linejoin" parameter value.
 **/
FoDocLineJoin
fo_doc_pdflib_get_line_join (FoDoc        *fo_doc)
{
  g_return_val_if_fail (FO_IS_DOC_PDFLIB (fo_doc), FO_DOC_LINE_JOIN_INVALID);

  return fo_doc->line_join;
}

/**
 * fo_doc_pdflib_set_line_join:
 * @fo_doc:   #FoDoc.
 * @line_join: New "linejoin" parameter value.
 * 
 * Set the "linejoin" parameter value of @fo_doc.
 **/
void
fo_doc_pdflib_set_line_join (FoDoc        *fo_doc,
			     FoDocLineJoin line_join)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);
  g_return_if_fail (line_join < FO_DOC_LINE_JOIN_LIMIT);

  /*if (line_join != fo_doc->line_join)
    {*/
      fo_doc->line_join = line_join;
      PDF_setlinejoin (FO_DOC_PDFLIB (fo_doc)->pdf_doc, line_join);
      /*}*/
}

/**
 * fo_doc_pdflib_get_line_width:
 * @fo_doc: #FoDoc.
 * 
 * Get the current line width of @fo_doc.
 * 
 * Return value: Current line width.
 **/
gfloat
fo_doc_pdflib_get_line_width (FoDoc        *fo_doc)
{
  g_return_val_if_fail (FO_IS_DOC_PDFLIB (fo_doc), 0);

  return fo_doc->line_width;
}

/**
 * fo_doc_pdflib_set_line_width:
 * @fo_doc:   #FoDoc.
 * @line_width: New line width value.
 * 
 * Set the line width of @fo_doc.
 **/
void
fo_doc_pdflib_set_line_width (FoDoc       *fo_doc,
			      gfloat       line_width)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  /*if (line_width != fo_doc->line_width)
    {*/
      fo_doc->line_width = line_width;
      PDF_setlinewidth (FO_DOC_PDFLIB (fo_doc)->pdf_doc, line_width);
      /*}*/
}

/**
 * fo_doc_pdflib_set_dash:
 * @fo_doc: #FoDoc.
 * @b:      Number of black units.
 * @w:      Number of white units.
 * 
 * Set the current dash pattern of @fo_doc.
 *
 * The dash pattern is the ratio of @b to @w.
 *
 * Set @b and @w to 0 to produce a solid line.
 *
 * The dash pattern is set to solid (@b = @w = 0) at the beginning of
 * each page.
 **/
void
fo_doc_pdflib_set_dash (FoDoc *fo_doc,
			gfloat b,
			gfloat w)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_setdash (FO_DOC_PDFLIB (fo_doc)->pdf_doc, b, w);
}

/**
 * fo_doc_pdflib_line_to:
 * @fo_doc: #FoDoc.
 * @x:      X-coordinate of the new current point.
 * @y:      Y-coordinate of the new current point.
 * 
 * Draw a line from the current point to another point.
 **/
void
fo_doc_pdflib_line_to (FoDoc *fo_doc,
		       gfloat x,
		       gfloat y)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_lineto (FO_DOC_PDFLIB (fo_doc)->pdf_doc, x, y);
}

/**
 * fo_doc_pdflib_translate:
 * @fo_doc: #FoDoc.
 * @x:      X-coordinate of the new origin of the coordinate system.
 * @y:      Y-coordinate of the new origin of the coordinate system.
 * 
 * Translate the origin of the coordinate system of @fo_doc.
 *
 * @x and @y are measured in the old coordinate system.
 **/
void
fo_doc_pdflib_translate (FoDoc *fo_doc,
			 gfloat x,
			 gfloat y)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_translate (FO_DOC_PDFLIB (fo_doc)->pdf_doc, x, y);
}

/**
 * fo_doc_pdflib_clip:
 * @fo_doc: #FoDoc.
 * 
 * Use the current path of @fo_doc as its clipping path.
 **/
void
fo_doc_pdflib_clip (FoDoc *fo_doc)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_clip (FO_DOC_PDFLIB (fo_doc)->pdf_doc);
}

/**
 * fo_doc_pdflib_save:
 * @fo_doc: #FoDoc.
 * 
 * Save the current graphics state of @fo_doc.
 **/
void
fo_doc_pdflib_save (FoDoc *fo_doc)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_save (FO_DOC_PDFLIB (fo_doc)->pdf_doc);
}

/**
 * fo_doc_pdflib_restore:
 * @fo_doc: #FoDoc.
 * 
 * Restore the most recently saved graphics state of @fo_doc.
 **/
void
fo_doc_pdflib_restore (FoDoc *fo_doc)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_restore (FO_DOC_PDFLIB (fo_doc)->pdf_doc);
}

/**
 * fo_doc_pdflib_move_to:
 * @fo_doc: #FoDoc.
 * @x:      X-coordinate of the new current point.
 * @y:      Y-coordinate of the new current point.
 * 
 * Set the current point of @fo_doc to (@x, @y).
 **/
void
fo_doc_pdflib_move_to (FoDoc *fo_doc,
		       gfloat x,
		       gfloat y)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_moveto (FO_DOC_PDFLIB (fo_doc)->pdf_doc, x, y);
}

/**
 * fo_doc_pdflib_line_stroked:
 * @fo_doc: #FoDoc.
 * @x0: X-coordinate of the start of the line.
 * @y0: Y-coordinate of the start of the line.
 * @x1: X-coordinate of the end of the line.
 * @y1: Y-coordinate of the end of the line.
 * 
 * Draw a line from (@x0,@y0) to (@x1,@y1).
 **/
void
fo_doc_pdflib_line_stroked (FoDoc *fo_doc,
			    gfloat x0,
			    gfloat y0,
			    gfloat x1,
			    gfloat y1)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_moveto (FO_DOC_PDFLIB (fo_doc)->pdf_doc, x0, y0);
  PDF_lineto (FO_DOC_PDFLIB (fo_doc)->pdf_doc, x1, y1);
}

/**
 * fo_doc_pdflib_rect_stroked:
 * @fo_doc: #FoDoc.
 * @x:      X-coordinate of the lower-left corner of the rectangle.
 * @y:      Y-coordinate of the lower-left corner of the rectangle.
 * @width:  Width of the rectangle.
 * @height: Height of the rectangle.
 * 
 * Draw an outline rectangle.
 **/
void
fo_doc_pdflib_rect_stroked (FoDoc *fo_doc,
			    gfloat x,
			    gfloat y,
			    gfloat width,
			    gfloat height)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_rect (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
	    x,
	    y,
	    width,
	    height);
  PDF_stroke (FO_DOC_PDFLIB (fo_doc)->pdf_doc);
}

/**
 * fo_doc_pdflib_rect_filled:
 * @fo_doc: #FoDoc.
 * @x:      X-coordinate of the lower-left corner of the rectangle.
 * @y:      Y-coordinate of the lower-left corner of the rectangle.
 * @width:  Width of the rectangle.
 * @height: Height of the rectangle.
 * 
 * Draw a filled rectangle.
 **/
void
fo_doc_pdflib_rect_filled (FoDoc *fo_doc,
		       gfloat x,
		       gfloat y,
		       gfloat width,
		       gfloat height)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_rect (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
	    x,
	    y,
	    width,
	    height);
  PDF_fill (FO_DOC_PDFLIB (fo_doc)->pdf_doc);
}

/**
 * fo_doc_pdflib_fill:
 * @fo_doc: #FoDoc.
 * 
 * Fill the interior of the path of @fo_doc with the current fill color.
 **/
void
fo_doc_pdflib_fill (FoDoc *fo_doc)
{
  g_return_if_fail (fo_doc != NULL);
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_fill (FO_DOC_PDFLIB (fo_doc)->pdf_doc);
}

/**
 * fo_doc_pdflib_stroke:
 * @fo_doc: #FoDoc.
 * 
 * Stroke the path of @fo_doc and clear the path.
 **/
void
fo_doc_pdflib_stroke (FoDoc *fo_doc)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_stroke (FO_DOC_PDFLIB (fo_doc)->pdf_doc);
}

/**
 * fo_doc_pdflib_fill_stroke:
 * @fo_doc: #FoDoc.
 * 
 * Fill and stroke the path of @fo_doc with the current fill and
 * stroke colors.
 **/
void
fo_doc_pdflib_fill_stroke (FoDoc *fo_doc)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_fill_stroke (FO_DOC_PDFLIB (fo_doc)->pdf_doc);
}

void
fo_doc_pdflib_place_image (FoDoc *fo_doc,
			   gint   image,
			   gfloat x,
			   gfloat y,
			   gfloat scale)
{
  PDF_place_image (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
		   image,
		   x,
		   y,
		   scale);
}

gint
fo_doc_pdflib_open_image_file (FoDoc      *fo_doc,
			       const char *imagetype,
			       const char *filename,
			       const char *stringparam,
			       gint        intparam)
{
  g_return_val_if_fail (FO_IS_DOC_PDFLIB (fo_doc), -1);
  g_return_val_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL, -1);

  return PDF_open_image_file (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
			      imagetype,
			      filename,
			      stringparam,
			      intparam);
}

void
fo_doc_pdflib_close_image (FoDoc *fo_doc,
			   gint   image)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);

  PDF_close_image (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
		   image);
}

/**
 * fo_doc_pdflib_render_layout_lines:
 * @fo_doc: 
 * @area_layout: 
 * @x: 
 * @y: 
 * 
 * 
 **/
void
fo_doc_pdflib_render_layout_lines (FoDoc   *fo_doc,
				   FoArea  *area_layout,
				   gfloat   x,
				   gfloat   y)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);
  g_return_if_fail (FO_IS_AREA_LAYOUT (area_layout));

  gfloat use_y = y;

  gint line_first = fo_area_layout_get_line_first (area_layout);
  if (line_first > 0)
    {
      use_y =
	y + fo_area_layout_get_line_height (area_layout,
					    line_first - 1);
    }

  pango_pdflib_render_layout_lines (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
				    fo_area_layout_get_layout (area_layout),
				    fo_area_layout_get_line_first (area_layout),
				    fo_area_layout_get_line_last (area_layout),
				    x * PANGO_SCALE,
				    use_y * PANGO_SCALE);
}

/**
 * fo_doc_pdflib_render_layout:
 * @fo_doc: 
 * @area_layout: 
 * @x: 
 * @y: 
 * 
 * 
 **/
void
fo_doc_pdflib_render_layout (FoDoc   *fo_doc,
			     FoArea  *area_layout,
			     gfloat   x,
			     gfloat   y)
{
  g_return_if_fail (FO_IS_DOC_PDFLIB (fo_doc));
  g_return_if_fail (FO_DOC_PDFLIB (fo_doc)->pdf_doc != NULL);
  g_return_if_fail (FO_IS_AREA_LAYOUT (area_layout));

  pango_pdflib_render_layout (FO_DOC_PDFLIB (fo_doc)->pdf_doc,
			      fo_area_layout_get_layout (area_layout),
			      x * PANGO_SCALE,
			      y * PANGO_SCALE);
}

/**
 * fo_doc_pdflib_debug_dump:
 * @object: #FoObject to be dumped.
 * @depth:  Relative indent to add to the output.
 * 
 * Implements #FoObject debug_dump method for #FoDocPDFLIB.
 **/
void
fo_doc_pdflib_debug_dump (FoObject *object,
			  gint      depth)
{
  FoDocPDFLIB *fo_doc_pdflib;
  gchar *indent = g_strnfill (depth * 2, ' ');
  gchar* object_sprintf;

  g_return_if_fail (object != NULL);
  g_return_if_fail (FO_IS_DOC_PDFLIB (object));

  fo_doc_pdflib = FO_DOC_PDFLIB (object);

  object_sprintf = fo_object_debug_sprintf (object);

  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%s%s",
	 indent,
	 object_sprintf);

  g_free (object_sprintf);

  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%s    pdf_doc: %p",
	 indent,
	 fo_doc_pdflib->pdf_doc);

  g_free (indent);

}
